#include <linux/module.h>
#include <linux/kernel.h> 
#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <linux/errno.h> 
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include "itecir.h"

MODULE_AUTHOR ("ITE,Inc.");
MODULE_DESCRIPTION ("ITE IT8712/IT8705 Consumer IR driver for Linux");
MODULE_LICENSE("Dual BSD/GPL");

MODULE_PARM(major,"i");
MODULE_PARM(irq,"i");

/* Global Variables */
int irq=4,major=0;
uioctl_t * pioc;
unsigned char tmpbuf[32];
unsigned char save[5];
//struct wait_queue *wq = NULL;
int signal=0 , m_newData = 0 ;
struct itecir_t {
	unsigned char Buf[32];
	unsigned char ModuFlag;
	unsigned char SyncFlag;
	unsigned char RXFreq;
	unsigned char RXRange;
	int BRate;
	unsigned char FreqApply;	
} *PIteCir;

void clear_fifo(void)
{
	unsigned int temp;
	temp=(inb(TCR1)|0x80);
	outb(temp,TCR1);
}
DECLARE_WAIT_QUEUE_HEAD(wq);
void ITE_CIR_INT(int irq, void *dev_id, struct pt_regs *regs)
{
	int temp,i;
	temp=(inb(IIR)&0x06);
	switch (temp) {
		case 0x2 : 
			break;
		case 0x4 :
			break;			
		case 0x6 :
			for (i=0;i<32;i++) {
				tmpbuf[i]=inb(DR);
			//	printk("%02X",tmpbuf[i]);
			//	if(i%8==7) printk("--");
			//	if(i%32==31) printk("\n");
			}
			clear_fifo();
			/* Receiver ACT */
			temp=(inb(RCR)|0x08);
			outb(temp,RCR);
			m_newData = 1 ;
			//wake_up_interruptible( &wq);
			break;
		default :
			break;
	}
}


void SetRXFace(void)
{
	unsigned int temp;
	outb(0x20,IER);
	/* set RCR */
	temp= (inb(RCR)&0x78) ; 	//disable HCFS
	temp|=PIteCir->SyncFlag <<7;    //set RDWOS
	temp|=PIteCir->RXRange;		//set RXDCR
	outb(temp,RCR);
	
	/* set BLDR and BHDR */
	temp=(inb(IER)|0x10);		//enable BR
	outb(temp,IER);
	outb(PIteCir->BRate,BLDR);	//after enable BR, set BLDR and BHDR
	outb(0x00,BHDR);
	temp &=(0xEF);			//disable BR
	outb(temp,IER);		

	/* enable all interrupt in IER */
	temp=(inb(IER)|0x0C);
	outb(temp,IER);

	/* set RCR_RXEND */
	temp=inb(RCR);
	temp|=(PIteCir->ModuFlag<<4 & 0xCF);
	outb(temp,RCR);
	clear_fifo();
}

int CIR_RX( uioctl_t * pioc ){
	if(m_newData)
	{
		int i;
		//printk(KERN_DEBUG"%i(%s)\n",current->pid,current->comm);
		//interruptible_sleep_on(&wq);// marked by brian
		for (i=0;i<32;i++){
			PIteCir->Buf[i]=tmpbuf[i];
		//	printk("%02X",PIteCir->Buf[i]);
		}
		copy_to_user(PIteCir,pioc->data,32);
		for (i=0;i<32;i++)
			tmpbuf[i]=0xFF;
		clear_fifo();
		m_newData = 0;
		return 0 ;
	}
	else
	{
		return -1;
	}	
}

void CIR_OPEN_RX( uioctl_t * pioc )
{
	unsigned char temp;       
	SetRXFace();
	clear_fifo();	
	/* enable RCR_RXEN*/
	temp=(inb(RCR)|0x20);
	outb(temp,RCR);
	
	/* enable RCR_RXACT*/
	temp=(inb(RCR)|0x08);
	outb(temp,RCR);
}

void CIR_TX( uioctl_t * pioc)
{
	unsigned char i;
	unsigned char temp;       
 
	SetRXFace();
	outb(0x80,TCR1);
	temp=(inb(RCR)&0x00);
	outb(temp,RCR);
	clear_fifo();
	
	for (i=0;i<32;i++) 
		outb(PIteCir->Buf[i],DR);
}

static int itedev_ioctl( struct inode * inode, struct file * filep, 
				unsigned int cmd, unsigned long	arg )
{
	/* Extract the type and number bitfield. */
	if (_IOC_TYPE(cmd) != ITE_IOCMAGIC )
		return -EINVAL;

	/* Allocate space for ioctl data structure. */
 	if ((pioc = kmalloc(sizeof(uioctl_t), GFP_KERNEL)) == NULL) {
     		printk("itedev_ioctl: error kmalloc on ioctl\n");
     		return -ENOMEM;
    	}

	/* Get the user ioctl structure. */
 	if (copy_from_user(pioc, (uioctl_t *)arg, sizeof(uioctl_t))) {
     		kfree(pioc);
		return -EFAULT;
	}
	
	PIteCir = pioc->data;
	/* 	 Check which command to do. */
 	switch (cmd)
	{
	case ITE_IOC_SEND:
		CIR_TX(pioc);
		return 0;
	case ITE_IOC_RECEIVE:
		if( CIR_RX(pioc) <0 )
			return -1;
		else 
			return 0 ;	
	case ITE_IOC_OPEN_RECEIVE:
		CIR_OPEN_RX(pioc);
		return 0;
	}
	return 0;
}

static int itedev_ioctl_entry ( struct inode *	inode, struct	file *	filep,
 	unsigned int cmd, unsigned long	arg)
{

 int ret = -1;
 ret = itedev_ioctl(inode, filep, cmd, arg);
 return ret;

} /* end itedev_ioctl_entry */

static int itedev_open(struct inode * inode, struct file * file)
{
 MOD_INC_USE_COUNT;
 return 0;
}

static int itedev_close(struct inode * inode, struct file * file)
{
 MOD_DEC_USE_COUNT;
 return 0;
}

struct file_operations itedev_fops = {
	ioctl: itedev_ioctl_entry ,
	open: itedev_open ,
        release: itedev_close ,
};




/* 
   above: driver API		
   below: driver init.
*/

void OutputData(unsigned char index,unsigned char value)
{
	outb(index,0x2e);
	outb(value,0x2f);
}


unsigned char InputData(unsigned char index)
{
	outb(index,0x2e);
	return inb(0x2f);
}

int GetChipID(void)
{
	unsigned int id1,id2;
	id1=InputData(0x20);
	id2=InputData(0x21);
	if (id1==0x87 && id2==0x05) 
		return 0x07;
	else if (id1==0x87 && id2==0x12) 
		return 0x0A;
	else 
		return -1;
}

int SetCIR(void)
{
	unsigned int LDN;
	int interval;
	LDN=GetChipID();
	if (LDN > 0)
		OutputData(0x07,LDN);		// Set LDN
	else {
		printk(KERN_INFO "Not ITE CHIP , EXIT...\n");
		return -ENODEV;
	}
	/* save the original value */
	save[0]=InputData(0x30);
	save[1]=InputData(0x60);
	save[2]=InputData(0x61);
	save[3]=InputData(0x70);
	save[4]=InputData(0xF0);

	OutputData(0x30,0x01); 			// Enable CIR
	OutputData(0x60,(iport >> 8));		// iport MSB
	OutputData(0x61,(iport & 0xFF));	// iport LSB
	OutputData(0x70,irq);			// Set IRQ
	OutputData(0xF0,0x00);			// Not Sharing IRQ

	if (check_region(0x310,8)) {
		printk("Error! ioport 0x310 has been used!");
		return -EBUSY;
	}
	else 
		request_region(0x310,8,"itecir");
		
	interval = request_irq(irq, ITE_CIR_INT, 0, "ITE_CIR", NULL);
	if (interval)
		return interval;
		
	major=register_chrdev(0, "itecir", &itedev_fops);
	return 0;
}

void ExitPnpMode(void)
{
	OutputData(0x02,0x02);
}

void EnterPnpMode(void)
{
	char str[4]={0x87,0x01,0x55,0x55};
	int i;
	for(i=0;i<4;i++)
		outb(str[i],0x2E);
}

static int __init ite_chip_init(void)
{
	int error;
	EnterPnpMode();
	error=SetCIR();
	if (error) {
		printk(KERN_INFO "ITE 8712/8705 SetCIR fail\n");
		return error;
	} else printk(KERN_INFO "ITE 8712/8705 CIR Loaded\n");
		
	ExitPnpMode();
		
	return 0;
}

static void __exit ite_chip_exit(void)
{
	/* restore original value*/
	EnterPnpMode();
	OutputData(0x30,save[0]);
	OutputData(0x60,save[1]);
	OutputData(0x61,save[2]);
	OutputData(0x70,save[3]);
	OutputData(0xF0,save[4]);
	ExitPnpMode();
	kfree(pioc);
	free_irq(irq,NULL);
	release_region(0x310,8);
	unregister_chrdev( major,"itecir");
	printk(KERN_INFO "ITE 8712/8705 CIR unload\n");
}

module_init(ite_chip_init);
module_exit(ite_chip_exit);

